/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package main;

import static opengl.GL.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.lwjgl.LWJGLException;
import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.Display;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;
import util.Util;

/**
 *
 * @author NMARNIOK
 */
public class CubeAndPyramid {
    private static int programID;       // shader program id
    private static int quadVAID;        // vertexarrayobject id of the cube
    private static int triangleVAID;    // vertexarrayobject id of the pyramid
    private static int modelLocation;   // uniform location of "model"
    private static final Vector2f rotation = new Vector2f();
    private static final Matrix4f cubeMat = new Matrix4f();
    private static final Matrix4f cubeModel[] = new Matrix4f[] {
        new Matrix4f(), new Matrix4f(), new Matrix4f(), new Matrix4f(),
    };
    private static final Matrix4f pyramidModel[] = new Matrix4f[] {
        new Matrix4f(), new Matrix4f(), new Matrix4f(), new Matrix4f(),
    };
    
    private static final Vector3f camPos = new Vector3f(0.0f, 0.0f, -4.0f);
    private static final Vector2f moveDir = new Vector2f(0.0f, 0.0f);
    private static boolean perspective = false;
    private static boolean bContinue = true;
    
    public static void main(String[] argv) {
        try {
            init();
            programID = Util.createShaderProgram("shader/Main_VS.glsl", "shader/VertexColor_FS.glsl");
            modelLocation = glGetUniformLocation(programID, "model");
            quadVAID = Util.createQuad();
            triangleVAID = Util.createTriangle();
            glPointSize(16.0f);
            glLineWidth(4.0f);
            initCube();
            initPyramid();
            render();
            Display.destroy();
        } catch (LWJGLException ex) {
            Logger.getLogger(CubeAndPyramid.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    public static void render() throws LWJGLException {
        glClearColor(0.0f, 0.0f, 0.25f, 1.0f); // dark blue
        long last = System.currentTimeMillis();
        long now;
        while(bContinue && !Display.isCloseRequested()) {            
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
            glUseProgram(programID);
            
            for(Matrix4f cubeSide : cubeModel) {
                Matrix4f temp = Util.mul(null, cubeMat, cubeSide);
                temp.store(Util.MAT_BUFFER);
                Util.MAT_BUFFER.position(0);
                glUniformMatrix4(modelLocation, false, Util.MAT_BUFFER);
                glBindVertexArray(quadVAID);
                glDrawArrays(GL_LINE_LOOP, 0, 4);
            }
            
            for(Matrix4f pyramidSide : pyramidModel) {
                pyramidSide.store(Util.MAT_BUFFER);
                Util.MAT_BUFFER.position(0);
                glUniformMatrix4(modelLocation, false, Util.MAT_BUFFER);
                glBindVertexArray(triangleVAID);
                glDrawArrays(GL_LINE_LOOP, 0, 3);
            }
            
            Display.update();
            Display.sync(60);
            
            now = System.currentTimeMillis();
            handleInput(now - last);
            updateUniforms();
            last = now;
        }
    }
    
    public static void initCube() {
        Matrix4f rot = Util.rotationY(Util.PI_DIV2, null);
        Util.mul(cubeModel[0], Util.translationX(-0.25f, null), rot, Util.scale(0.25f, null));
        Util.mul(cubeModel[1], Util.translationZ(+0.25f, null),      Util.scale(0.25f, null));
        Util.mul(cubeModel[2], Util.translationX(+0.25f, null), rot, Util.scale(0.25f, null));
        Util.mul(cubeModel[3], Util.translationZ(-0.25f, null),      Util.scale(0.25f, null));
    }
    
    public static void initPyramid() {
        Matrix4f tilt = Util.rotationX(Util.PI / 6.0f, null);
        Matrix4f rot = Util.rotationY(Util.PI_DIV2, null);
        Matrix4f translation = Util.translationX(-0.5f, null);
        Util.mul(pyramidModel[0], translation, Util.translationX(-0.125f, null), rot,           tilt, Util.scale(0.25f, null));
        Util.mul(pyramidModel[1], translation, Util.translationZ(+0.125f, null), rot, rot,      tilt, Util.scale(0.25f, null));
        Util.mul(pyramidModel[2], translation, Util.translationX(+0.125f, null), rot, rot, rot, tilt, Util.scale(0.25f, null));
        Util.mul(pyramidModel[3], translation, Util.translationZ(-0.125f, null),                tilt, Util.scale(0.25f, null));
    }
    
    public static void handleInput(long millis) {
        float rotSpeed = 5e-3f*(Keyboard.isKeyDown(Keyboard.KEY_RSHIFT) ? 2.0f : 1.0f)*(float)millis;
        float moveSpeed = 2e-3f*(Keyboard.isKeyDown(Keyboard.KEY_LSHIFT) ? 2.0f : 1.0f)*(float)millis;
        float camSpeed = 5e-3f;
        
        while(Keyboard.next()) {
            if(Keyboard.getEventKeyState()) {
                switch(Keyboard.getEventKey()) {
                    case Keyboard.KEY_W: moveDir.y += 1.0f; break;
                    case Keyboard.KEY_S: moveDir.y -= 1.0f; break;
                    case Keyboard.KEY_A: moveDir.x += 1.0f; break;
                    case Keyboard.KEY_D: moveDir.x -= 1.0f; break;
                }
            } else {
                switch(Keyboard.getEventKey()) {
                    case Keyboard.KEY_W: moveDir.y -= 1.0f; break;
                    case Keyboard.KEY_S: moveDir.y += 1.0f; break;
                    case Keyboard.KEY_A: moveDir.x -= 1.0f; break;
                    case Keyboard.KEY_D: moveDir.x += 1.0f; break;
                    case Keyboard.KEY_F1: perspective ^= true; break;
                }
            }
        }
        
        camPos.x += moveSpeed * moveDir.x;
        camPos.y += moveSpeed * moveDir.y;
        camPos.z = 1.0f;
        
        while(Mouse.next()) {
            if(Mouse.getEventButton() == 0) {
                Mouse.setGrabbed(Mouse.getEventButtonState());
            }
            if(Mouse.isGrabbed()) {                
                camPos.x -= camSpeed*Mouse.getEventDX();
                camPos.y -= camSpeed*Mouse.getEventDY();
            }
        }
        
        Util.clamp(camPos, -1.0f, +1.0f, camPos);
        
        if(Keyboard.isKeyDown(Keyboard.KEY_UP)) rotation.x += rotSpeed;
        if(Keyboard.isKeyDown(Keyboard.KEY_DOWN)) rotation.x -= rotSpeed;
        if(Keyboard.isKeyDown(Keyboard.KEY_LEFT)) rotation.y += rotSpeed;
        if(Keyboard.isKeyDown(Keyboard.KEY_RIGHT)) rotation.y -= rotSpeed;
        if(Keyboard.isKeyDown(Keyboard.KEY_ESCAPE)) bContinue = false;
    }
    
    public static void updateUniforms() {        
        Matrix4f cubeTrans = Util.translationX(+0.5f, null);
        Util.mul(cubeMat, cubeTrans, Util.rotationX(rotation.x, null), Util.rotationY(rotation.y, null));
                
        Matrix4f view = Util.lookAtRH(camPos, new Vector3f(), new Vector3f(0.0f, 1.0f, 0.0f), null);
        Matrix4f projection = new Matrix4f();
        if(perspective) {
            Util.perspectiveRH(-1e-2f, 1e-2f, -1e-2f, 1e-2f, 1e-2f, 1e+2f, projection);
        } else {
            Util.orthographicRH(projection);
        }
        Matrix4f viewProj = Matrix4f.mul(projection, view, null);
        viewProj.store(Util.MAT_BUFFER);
        Util.MAT_BUFFER.position(0);
        int viewProjLoc = glGetUniformLocation(programID, "viewProj");
        glUniformMatrix4(viewProjLoc, false, Util.MAT_BUFFER);
    }
}
